#include <stdio.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <fcntl.h>
char *strdup(), *strchr();
typedef unsigned char  uint1;
typedef unsigned int   uint4;

#ifdef STRERROR
extern char *sys_errlist[];
#define strerror(n) sys_errlist[n]
#endif

#define ARC_REV_TABSIZE 109121

#define ch_term(t) ((t=='<')||(t=='>')||(t=='(')||(t==')')||(t=='"')||(t==',')||(t<=32))
#define ch_num(c) ((c>='0')&&(c<='9'))
#define ch_alpha(c) (((c>='a')&&(c<='z'))||((c>='A')&&(c<='Z')))

FILE *jfopen(char *path, char *mode)
{
  FILE *res = fopen(path, mode);
  if (res==0) {
    fprintf(stderr,"Could not open %s for %s\n", path, mode);
    exit(1);
  }
  return res;
}

FILE *jpopen(char *path, char *mode)
{
  FILE *res = popen(path, mode);
  if (res==0) {
    fprintf(stderr,"Could not popen %s for %s\n", path, mode);
    exit(1);
  }
  return res;
}

void jfseek(FILE *f, int pos, int rel)
{
  int ok;
  fflush(f);
  ok = fseek(f, pos, rel);
  if (ok < 0) {
    fprintf(stderr,"seek: %s\n", strerror(errno));
    exit(1);
  }
  fflush(f);
}

void jfread(void *ptr, int size, int nitems, FILE *f)
{
  int res = fread(ptr, size, nitems, f);
  if (res != nitems) {
    fprintf(stderr,"data file corrupted.\n");
    exit(1);
  }
}

void jfwrite(void *ptr, int size, int nitems, FILE *f)
{
  int res = fwrite(ptr, size, nitems, f);
  if (res != nitems) {
    fprintf(stderr,"cannot write output file.\n");
    exit(1);
  }
}

void jwriteint(int n, FILE *f)
{
  putc((n>>24)&0xFF, f);
  putc((n>>16)&0xFF, f);
  putc((n>>8)&0xFF, f);
  putc((n)&0xFF, f);
}

int jreadint(FILE *f)
{
  unsigned char c0 = getc(f);
  unsigned char c1 = getc(f);
  unsigned char c2 = getc(f);
  unsigned char c3 = getc(f);
  return (c0<<24)+(c1<<16)+(c2<<8)+c3;
}

void jreadword(FILE *f, char *buf, int max)
{
  int pos = 0; int c;
  while (1) {
    c=getc(f);
    if ((c==EOF)||(c==' ')) break;
    if (pos<max) buf[pos++] = c;
  }
  buf[pos]=0;
}

void jwriteintpk(unsigned int n, FILE *f)
{
  if (n<=0x3F) { 
    putc(n, f); return;
  }
  if (n<=0x3FFF) { 
    putc((n>>8)|0x40,f); putc(n&0xFF, f); return;
  }
  if (n<=0x3FFFFF) { 
    putc((n>>16)|0x80,f); putc((n>>8)&0xFF, f);
    putc(n&0xFF,f); return;
  }
  if (n<=0x3FFFFFFF) {
    putc((n>>24)|0xc0,f); putc((n>>16)&0xFF, f); 
    putc((n>>8)&0xFF, f); putc((n&0xFF), f); return;
  }
  fprintf(stderr,"Cannot write data, too large.\n");
  exit(1);
}

unsigned int jreadintpk(FILE *f)
{
  unsigned int b0,b1,b2,b3;
  b0 = getc(f);
  switch (b0&0xC0) {
  case 0x00: 
    return b0;
  case 0x40: 
    b1=getc(f); 
    return ((b0&0x3F)<<8)|b1;
  case 0x80: 
    b1=getc(f); b2=getc(f); 
    return ((b0&0x3F)<<16)|(b1<<8)|b2;
  case 0xc0:
    b1=getc(f); b2=getc(f); b3=getc(f);
    return ((b0&0x3f)<<24)|(b1<<16)|(b2<<8)|b3;
  }
}

char *strclip(char *p, char *e)
{
  int len = e-p;
  char *res = (char *)malloc(len+1);
  memcpy(res, p, len);
  res[len] = 0;
  return res;
}

void downcase(char *s)
{
  while (*s) {
    char c = *s;
    if ((c>='A')&&(c<='Z')) *s = c+32;
    s++;
  }
}

int strhash(char *p, int modulus)
{
  int hashval = 0;
  while (*p) {
    hashval = (hashval<<11)|(hashval>>21);
    hashval ^= *p; p++;
  }
  hashval = (hashval&0x7FFFFFFF) % modulus;
  return hashval;
}

char *field(int index, char *str)
{
  while (index) { 
    if (str==0) return 0;
    str = strchr(str, ',')+1;
    index--;
  }
  return str;
}

void htmlputc(char c)
{
  if (((c>='a')&&(c<='z'))||
      ((c>='A')&&(c<='Z'))||
      ((c>='0')&&(c<='9'))) {
    putchar(c);
  } else if (c<=32) {
    if (c=='\n') printf("<BR>\n");
    else putchar(' ');
  } else printf("&#%d;",c);
}

void htmlputs(char *s)
{
  char *p; char c;
  while (*s) {
    p = s;
    if (strncasecmp(p,"http://",7)==0) {
      p += 7;
      while ((*p>32)&&(*p!='"')&&(*p!='>')) p++;
      c=*p; *p=0;
      printf("<A HREF=\"%s\">%s</A>",s,s);
      *p=c;
      s=p; continue;
    } else htmlputc(*s++);
  }
}

void htmlfield(int n, char *s)
{
  s = field(n, s);
  while ((*s)&&(*s!=',')&&(*s!='\n')) htmlputc(*s++);
}

/******************************************************************************
 *
 * message_string
 *
 * - converts a message which has been decomposed into its parts
 *   back into a single string.
 *
 *****************************************************************************/

char *message_string(char *from, char **heads, char **bodys)
{
  int i, len; char *res;
  len = strlen(from) + 1;
  for (i=0; heads[i]; i++) len+=strlen(heads[i])+1;
  for (i=0; bodys[i]; i++) len+=strlen(bodys[i])+1;
  len+=1;
  res = (char *)malloc(len+1);
  len = 0;
  strcpy(res+len, from); 
  len+=strlen(res+len);
  res[len++] = '\n';
  for (i=0; heads[i]; i++) {
    strcpy(res+len, heads[i]);
    len += strlen(res+len);
    res[len++] = '\n';
  }
  res[len++] = '\n';
  for (i=0; bodys[i]; i++) {
    strcpy(res+len, bodys[i]);
    len += strlen(res+len);
    res[len++] = '\n';
  }
  res[len] = 0;
  return res;
}

/******************************************************************************
 *
 * intset operations.
 *
 * - these are currently represented as vectors of integers,
 *   always sorted.
 *
 *****************************************************************************/

typedef struct intset
{
  int size;
  int data[1];
}
*intset;

struct intset intset_none = { 0, {0} };

#define INTSET_NONE  (&intset_none)
#define INTSET_ALL   0

intset intset_make(int n)
{
  intset res = (intset)malloc((1+n)*sizeof(int));
  res->size = n;
  return res;
}

intset intset_union(intset *vectors, int nvectors)
{
  int i, j, min, max, total, dcount; intset v;
  unsigned char *counts;
  min = 0x7FFFFFFF; max = 0;
  for (i=0; i<nvectors; i++) {
    v = vectors[i];
    if (v == INTSET_ALL) continue;
    if (v->size == 0) return INTSET_NONE;
    for (j=0; j<v->size; j++) {
      if (v->data[j] < min) min = v->data[j];
      if (v->data[j] > max) max = v->data[j];
    }
  }
  max++;
  if (min>=max) return INTSET_ALL;
  if (max>min+1000000) {
    fprintf(stderr,"Bitvector overflow.\n");
    exit(1);
  }
  counts = (unsigned char *)malloc(max-min);
  memset(counts, 0, max-min); dcount=0;
  for (i=0; i<nvectors; i++) {
    v = vectors[i];
    if (v == INTSET_ALL) continue;
    dcount++;
    for (j=0; j<v->size; j++) {
      int n = v->data[j];
      counts[n-min]++;
    }
  }
  total = 0;
  for (i=0; i<(max-min); i++)
    if (counts[i]==dcount)
      total++;
  if (total==0) return INTSET_NONE;
  v = intset_make(total); j=0;
  for (i=0; i<(max-min); i++)
    if (counts[i]==dcount)
      v->data[j++] = (i+min);
  return v;
}

void intset_free(intset s)
{
  free(s);
}

/******************************************************************************
 *
 * static huffman coding.
 *
 * - This module implements static huffman coding.  The static table is
 *   optimized for the English language.  It typically obtains compression
 *   ratios of about xxxx on English text.
 *
 * int huff_file_init(FILE *f)
 *
 * - The specified file is cleared out and initialized as a huffman
 *   file.  Huffman files are basically strings stored in a compact
 *   format.
 *
 * int huff_file_append(FILE *f, char *src, int srcbytes)
 *
 * - Appends the specified string to the given huffman file.
 *   Returns a file-position which can be used to seek to the
 *   start of the newly-appended string.
 *
 * int huff_file_read(FILE *f, int pos, char *buf, int buflen)
 *
 * - Read the specified number of bytes from the huffman file at
 *   the specified position.  Returns the number of bytes actually
 *   read.
 *
 *****************************************************************************/

uint1 huff_ch_nbits[256]={
16, 16, 16, 16, 16, 16, 16, 16, 16, 11,  5, 16, 16, 16, 16, 16, 
16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 
 3, 10,  7, 11, 14, 11, 15,  8, 10,  9, 12, 12,  6,  7,  6,  7, 
 8,  8,  8,  9,  9,  9,  9, 10, 10,  8,  8, 10,  9, 12,  9,  9, 
 9,  8,  9,  8,  9,  9,  9, 10,  9,  8,  9,  9,  9,  8,  9,  9, 
10, 12,  8,  8,  8, 10, 11,  8, 10,  9, 10,  9, 15, 13, 16,  7, 
14,  4,  7,  6,  5,  4,  6,  6,  5,  5,  9,  7,  5,  6,  4,  4, 
 6, 11,  5,  5,  4,  5,  7,  6,  9,  6, 10, 13, 16, 13, 14, 16, 
16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 
16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 
16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 
16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 
16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 
16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 
16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 
16, 16, 16, 16, 16, 15, 15, 15, 15, 15, 14, 14, 14, 14, 14, 14, 
};

uint4 huff_ch_bits[256]={
0xfdff, 0xfe76, 0xfe77, 0xfe7d, 0xfe7e, 0xfe7f, 0xfed0, 0xfed1, 
0xfed2, 0x07e7, 0x0015, 0xfed3, 0xfed4, 0xfed5, 0xfed6, 0xfed7, 
0xfed8, 0xfed9, 0xfeda, 0xfedb, 0xfedc, 0xfedd, 0xfede, 0xfedf, 
0xfee0, 0xfee1, 0xfee2, 0xfee3, 0xfee4, 0xfee5, 0xfee6, 0xfee7, 
0x0000, 0x03ee, 0x006e, 0x07e6, 0x3f9c, 0x07f2, 0x7efe, 0x00ec, 
0x03ec, 0x01f3, 0x0fe6, 0x0fd6, 0x0034, 0x006c, 0x0030, 0x0070, 
0x00e6, 0x00e4, 0x00eb, 0x01ec, 0x01e1, 0x01f2, 0x01e0, 0x03f2, 
0x03f4, 0x00e3, 0x00e2, 0x03fa, 0x01fe, 0x0fde, 0x01e3, 0x01e9, 
0x01f4, 0x00e5, 0x01e4, 0x00e7, 0x01e2, 0x01f0, 0x01e8, 0x03f8, 
0x01ea, 0x00e8, 0x01e5, 0x01eb, 0x01ef, 0x00ed, 0x01f5, 0x01f1, 
0x03f0, 0x0fec, 0x00e9, 0x00ee, 0x00ea, 0x03ed, 0x07ee, 0x00ef, 
0x03ef, 0x01e7, 0x03f6, 0x01ed, 0x7f3a, 0x1faf, 0xfdfe, 0x006b, 
0x3f9e, 0x0005, 0x006a, 0x002c, 0x0011, 0x0002, 0x002d, 0x0032, 
0x0014, 0x000e, 0x01ee, 0x006d, 0x000f, 0x002e, 0x0006, 0x0004, 
0x0031, 0x07ea, 0x0010, 0x0012, 0x0003, 0x0013, 0x006f, 0x0033, 
0x01e6, 0x002f, 0x03f1, 0x1fae, 0xfe7c, 0x1fbe, 0x3f7e, 0xfee8, 
0xfee9, 0xfeea, 0xfeeb, 0xfeec, 0xfeed, 0xfeee, 0xfeef, 0xfef0, 
0xfef1, 0xfef2, 0xfef3, 0xfef4, 0xfef5, 0xfef6, 0xfef7, 0xfef8, 
0xfef9, 0xfefa, 0xfefb, 0xfefc, 0xfefd, 0xfefe, 0xfeff, 0xff80, 
0xff81, 0xff82, 0xff83, 0xff84, 0xff85, 0xff86, 0xff87, 0xff88, 
0xff89, 0xff8a, 0xff8b, 0xff8c, 0xff8d, 0xff8e, 0xff8f, 0xff90, 
0xff91, 0xff92, 0xff93, 0xff94, 0xff95, 0xff96, 0xff97, 0xff98, 
0xff99, 0xff9a, 0xff9b, 0xff9c, 0xff9d, 0xff9e, 0xff9f, 0xffa0, 
0xffa1, 0xffa2, 0xffa3, 0xffa4, 0xffa5, 0xffa6, 0xffa7, 0xffa8, 
0xffa9, 0xffaa, 0xffab, 0xffac, 0xffad, 0xffae, 0xffaf, 0xffb0, 
0xffb1, 0xffb2, 0xffb3, 0xffb4, 0xffb5, 0xffb6, 0xffb7, 0xffb8, 
0xffb9, 0xffba, 0xffbb, 0xffbc, 0xffbd, 0xffbe, 0xffbf, 0xffc0, 
0xffc1, 0xffc2, 0xffc3, 0xffc4, 0xffc5, 0xffc6, 0xffc7, 0xffc8, 
0xffc9, 0xffca, 0xffcb, 0xffcc, 0xffcd, 0xffce, 0xffcf, 0xffd0, 
0xffd1, 0xffd2, 0xffd3, 0xffd4, 0xffd5, 0xffd6, 0xffd7, 0xffd8, 
0xffd9, 0xffda, 0xffdb, 0xffdc, 0xffdd, 0x7fef, 0x7ff0, 0x7ff1, 
0x7ff2, 0x7ff3, 0x3ffa, 0x3ffb, 0x3ffc, 0x3ffd, 0x3ffe, 0x3fff, 
};

uint1 *huff_bits_ch;

int huff_init()
{
  unsigned int i, j, bits, nbits, lo, hi;
  huff_bits_ch = (unsigned char *)malloc(65536);
  for (i=0; i<256; i++) {
    bits = huff_ch_bits[i];
    nbits = huff_ch_nbits[i];
    lo = (bits<<(16-nbits));
    hi = ((bits+1)<<(16-nbits));
    for (j=lo; j<hi; j++) huff_bits_ch[j] = i;
  }
}

void huff_file_init(FILE *f)
{
  jfseek(f, 0, 0);
  putc(0, f);
  putc(0, f);
  putc(0, f);
  putc(0, f);
  putc(0, f);
  putc(0, f);
}

int huff_file_append(FILE *f, char *src, int srcbytes)
{
  uint4 tpos, bits, nbits, queue, queuelen, result; uint1 c;
  fflush(f);
  jfseek(f, 0, 0);
  tpos  = getc(f); tpos<<=8;
  tpos |= getc(f); tpos<<=8;
  tpos |= getc(f); tpos<<=8;
  tpos |= getc(f);
  jfseek(f, 4+(tpos>>3), 0);
  queuelen = (tpos&7); queue = 0;
  if (queuelen) queue=(getc(f)<<24);
  jfseek(f, 4+(tpos>>3), 0);
  fflush(f);
  result = tpos;
  while (srcbytes) {
    c = *src++; srcbytes--;
    bits = huff_ch_bits[c];
    nbits = huff_ch_nbits[c];
    bits <<= (32 - nbits - queuelen);
    queue |= bits;
    queuelen += nbits;
    while (queuelen >= 8) { putc((queue>>24), f); queue<<=8; queuelen-=8; }
    tpos += nbits;
  }
  if (queuelen) putc((queue>>24),f);
  putc(0, f);
  putc(0, f);
  fflush(f);
  jfseek(f, 0, 0);
  putc((tpos>>24)&0xFF, f);
  putc((tpos>>16)&0xFF, f);
  putc((tpos>>8)&0xFF, f);
  putc((tpos>>0)&0xFF, f);
  fflush(f);
  return result;
}

int huff_file_read(FILE *f, int pos, char *buf, int buflen, int *npos)
{
  uint4 tpos, queue, queuelen, nread, val, c, nbits;
  if (huff_bits_ch==0) huff_init();
  fflush(f);
  jfseek(f, 0, 0);
  tpos  = getc(f); tpos<<=8;
  tpos |= getc(f); tpos<<=8;
  tpos |= getc(f); tpos<<=8;
  tpos |= getc(f);
  nread = 0;
  jfseek(f, 4+(pos>>3), 0);
  queue = (getc(f)<<24);
  queue <<= (pos&7);
  queuelen = (8-(pos&7));
  while ((nread < buflen)&&(pos < tpos)) {
    while (queuelen<16) { queuelen+=8; queue|=(getc(f) << (32-queuelen)); }
    val  = (queue>>16);
    c = huff_bits_ch[val & 0xFFFF];
    *buf++ = c; nread++;
    nbits = huff_ch_nbits[c];
    queuelen-=nbits;
    queue<<=nbits;
    pos += nbits;
  }
  if (npos) *npos = pos;
  return nread;
}

/******************************************************************************
 *
 * Header parsing.
 *
 * parse_headers(char **headers, 
 *		  char *name, char *email, char *subj, int *date)
 *
 * Given a vector of headers, extracts the name from the from-header,
 * the email-address from the from-header, the subject from the
 * subject-header, and the date from the date-header.
 *
 * The parse routine will remove all commas from the specified
 * fields, and will truncate the fields to 100 chars each.
 *
 *****************************************************************************/

void parse_from(char *from, char *user, char *email)
{
  char buffer[1000]; char *p, *e, *d, c;
  strncpy(buffer,from,999); buffer[999]=0;
  p = strchr(buffer,'@');
  if (p) {
    e = p;
    while ((p>buffer)&&(!ch_term(*p))) p--;
    if (ch_term(*p)) p++;
    while (!ch_term(*e)) e++;
    d = e; while (d-p > 99) d--;
    memcpy(email, p, d-p);
    email[d-p] = 0;
    while (p<e) *p++ =' ';
  }
  p = user;
  d = buffer+5;
  e = user+99;
  while (*d) {
    c = *d++;
    if (p<e) {
      if (ch_term(c)) {
	if ((p>user)&&(p[-1]!=' ')) *p++ = ' ';
      } else *p++ = c;
    }
  }
  while ((p>user)&&(p[-1]<=32)) p--;
  *p = 0;
  if (user[0]==0) {
    strcpy(user, email);
    p = strchr(user,'@');
    if (p) *p=0;
  }
}

void parse_subj(char *subj, char *result)
{
  int len, i;
  subj += 8;
strip:
  if (*subj==' ') { subj++; goto strip; }
  if (strncmp(subj,"Re:",3)==0) { subj+=3; goto strip; }
  if (strncmp(subj,"RE:",3)==0) { subj+=3; goto strip; }
  if (strncmp(subj,"re:",3)==0) { subj+=3; goto strip; }
  len = strlen(subj);
  while ((len>0)&&(subj[len-1]<=32)) len--;
  if (len>99) len=99;
  memcpy(result, subj, len);
  result[len] = 0;
  for (i=0; i<len; i++) if (result[len]==',') result[len]=';';
}

void parse_date(char *date, int *result)
{
  char *p, *e; int len; struct tm t;
  p = date + 5;
  t.tm_min = t.tm_hour = t.tm_mday = t.tm_mon = t.tm_year = -1;
  t.tm_wday = t.tm_yday = t.tm_isdst = t.tm_sec = 0;
  while (1) {
    while ((*p==' ')||(*p=='\t')) p++;
    if ((*p=='\n')||(*p=='\r')||(*p==0)) break;
    e=p;
    while ((*e!=' ')&&(*e!='\t')&&(*e!='\n')&&(*e!='\r')&&(*e!=0)) e++;
    len = e-p;
    /* Check for month-names */
    if ((len == 3)&&(*p>='A')&&(*p<='Z')) {
      if (memcmp(p,"Jan",3)==0) t.tm_mon = 0;
      if (memcmp(p,"Feb",3)==0) t.tm_mon = 1;
      if (memcmp(p,"Mar",3)==0) t.tm_mon = 2;
      if (memcmp(p,"Apr",3)==0) t.tm_mon = 3;
      if (memcmp(p,"May",3)==0) t.tm_mon = 4;
      if (memcmp(p,"Jun",3)==0) t.tm_mon = 5;
      if (memcmp(p,"Jul",3)==0) t.tm_mon = 6;
      if (memcmp(p,"Aug",3)==0) t.tm_mon = 7;
      if (memcmp(p,"Sep",3)==0) t.tm_mon = 8;
      if (memcmp(p,"Oct",3)==0) t.tm_mon = 9;
      if (memcmp(p,"Nov",3)==0) t.tm_mon = 10;
      if (memcmp(p,"Dec",3)==0) t.tm_mon = 11;
    }
    /* Check for single-digit month-numbers */
    if ((len == 1)&&ch_num(p[0])) {
      t.tm_mday = atoi(p);
    }
    /* Check for double-digit numbers */
    if ((len == 2)&&ch_num(p[0])&&ch_num(p[1])) {
      if (t.tm_mday>=0) t.tm_year = atoi(p);
      else t.tm_mday = atoi(p);
    }
    /* Check for HH:MM */
    if ((len == 5)&&(p[2]==':')) {
      t.tm_hour = atoi(p);
      t.tm_min = atoi(p+3);
    }
    /* Check for HH:MM:SS */
    if ((len == 8)&&(p[2]==':')&&(p[5]==':')) {
      t.tm_hour = atoi(p);
      t.tm_min = atoi(p+3);
      t.tm_sec = atoi(p+6);
    }
    /* Check for 19xx */
    if ((len==4)&&(p[0]=='1')&&(p[1]=='9')) {
      t.tm_year = atoi(p)-1900;
    }
    /* Check for 20xx */
    if ((len==4)&&(p[0]=='2')&&(p[1]=='0')) {
      t.tm_year = atoi(p)-1900;
    }
    p = e;
  }
  if (t.tm_year<0) { *result=0; return; }
  if (t.tm_mon<0) { *result=0; return; }
  if (t.tm_mday<0) { *result=0; return; }
  if (t.tm_hour<0) { *result=0; return; }
  if (t.tm_min<0) { *result=0; return; }
  *result = mktime(&t);
}

void parse_header(char **headers,
		  char *user, char *email, char *subj, int *date)
{
  int i; char *head;
  *date = 0; email[0]=0; user[0]=0; subj[0]=0;
  for (i=0; headers[i]; i++) {
    head = headers[i];
    if (strncmp(head,"From:",5)==0) parse_from(head, email, user);
    if (strncmp(head,"Subject:",8)==0) parse_subj(head, subj);
    if (strncmp(head,"Date:",5)==0) parse_date(head, date);
  }
}

/******************************************************************************
 *
 * Line classification
 *
 *****************************************************************************/

int match(char *line, char *template)
{
  while (1) {
    switch (*template++) {
    case 0: return (*line==0);
    case '*': 
      if (*line==' ') return 0;
      while ((*line)&&(*line!=' ')) line++; break;
    case ' ': 
      if (*line!=' ') return 0; 
      while ((*line)&&(*line==' ')) line++; break;
    case '+': if ((*line<'A')||(*line>'Z')) return 0; line++; break;
    case '-': if ((*line<'a')||(*line>'z')) return 0; line++; break;
    case '0': if ((*line<'0')||(*line>'9')) return 0; line++; break;
    case ':': if (*line!=':') return 0; line++; break;
    }
  }
}

int line_isblank(char *line)
{
  return (*line)==0;
}

int line_isnull(char *line)
{
  while (*line) {
    if (((*line>='a')&&(*line<='z'))||
	((*line>='A')&&(*line<='Z'))||
	((*line>='0')&&(*line<='9'))) return 0;
    line++;
  }
  return 1;
}

int line_isrfc(char *line)
{
  if (((*line<'A')||(*line>'Z'))&&
      ((*line<'a')||(*line>'z'))) return 0;
  line++;
  while (((*line>='A')&&(*line<='Z'))||
	 ((*line>='a')&&(*line<='z'))||
	 ((*line>='0')&&(*line<='9'))||
	 (*line=='-')) line++;
  if (*line!=':') return 0;
  return 1;
}

int line_isfrom(char *line)
{
  if (strncmp(line, "From ", 5)) return 0;
  line += 4;
  if (match(line," * +-- +-- * 00:00:00 0000")) return 1;
  if (match(line," * +-- +-- * 00:00 +++ 0000")) return 1;
  if (match(line," * +-- +-- * 00:00 0000")) return 1;
  if (match(line," * +-- +-- * 00:00:00 +++ 0000")) return 1;
  return 0;
}

/******************************************************************************
 *
 * splitmail
 *
 * This subroutine divides a UNIX mailbox file into mail messages.
 * In doing so, it attempts to correct errors in the format of the
 * mail file.
 *
 *****************************************************************************/

typedef void (*splitmail_fn)(char *, char **, char **);

int splitmail_getline(FILE *f, char *buffer, int buflen)
{
  int len;
  buffer[0]=0;
  if (fgets(buffer, buflen, f)==0) return -1;
  len = strlen(buffer);
  while ((len>0)&&(buffer[len-1]<=' ')) len--;
  buffer[len] = 0;
  return 0;
}

void splitmail_syntax()
{
  fprintf(stderr,"Input is not a correct UNIX mailbox file.\n");
  exit(1);
}

void splitmail(FILE *f, splitmail_fn fn)
{
  char line[10000]; int ok;
  char *from;
  char **heads; int nheads=0;
  char **bodys; int nbodys=0;

  heads = (char **)malloc(10000*sizeof(char*));
  bodys = (char **)malloc(1000000*sizeof(char*));
  ok = splitmail_getline(f, line, 9999);
  if (ok<0) return;
  if (!line_isfrom(line)) splitmail_syntax();
  from = strdup(line);
  while (1) {
    while (1) {
      ok = splitmail_getline(f, line, 9999);
      if (ok<0) splitmail_syntax();
      if (line[0]==0) break;
      if (nheads>=9999) {
	fprintf(stderr,"Too many mail headers.  Truncating.\n");
      } else heads[nheads++] = strdup(line);
    }
    while (1) {
      ok = splitmail_getline(f, line, 9999);
      if (ok<0) break;
      if (line_isfrom(line)) break;
      if (nbodys>=999999) {
	fprintf(stderr,"Mail message too big.  Truncating.\n");
      } else bodys[nbodys++] = strdup(line);
    }
    heads[nheads] = 0;
    bodys[nbodys] = 0;
    fn(from, heads, bodys);
    if (ok<0) break;
    from = strdup(line);
    nheads = 0;
    nbodys = 0;
  }
}

/******************************************************************************
 *
 * keep-headers
 *
 * removes all headers except the specified ones.
 *
 *****************************************************************************/

char **keep_headers_fields;

void keep_headers_handler(char *from, char **heads, char **bodys)
{
  int i, j, len; char *head;
  printf("%s\n", from); 
  free(from);
  for (i=0; heads[i]; i++) {
    head = heads[i];
    for (j=0; keep_headers_fields[j]; j++) {
      len = strlen(keep_headers_fields[j]);
      if ((head[len]==':')&&(strncasecmp(head, keep_headers_fields[j], len)==0)) {
	printf("%s\n", head);
	break;
      }
    }
    free(head);
  }
  printf("\n");
  for (i=0; bodys[i]; i++) {
    printf("%s\n", bodys[i]);
    free(bodys[i]);
  }
}

void keep_headers(char **argv)
{
  keep_headers_fields = argv;
  splitmail(stdin, keep_headers_handler);
}

/******************************************************************************
 *
 * divide
 *
 *****************************************************************************/

char *divide_strings[1000];
FILE *divide_files[1000];
int   divide_count;
FILE *divide_default;

int divide_match(char *header, char *pattern)
{
  char pc, hc;
  while (1) {
    hc = *header++;
    pc = *pattern++;
    if (pc==0) return 1;
    if (hc==0) return 0;
    if ((pc==' ')||(pc=='\t')) {
      if ((hc!=' ')&&(hc!='\t')) return 0;
      while ((*header==' ')||(*header=='\t')) header++;
      while ((*pattern==' ')||(*pattern=='\t')) pattern++;
      continue;
    }
    if ((hc>='A')&&(hc<='Z')) hc+=32;
    if ((pc>='A')&&(pc<='Z')) pc+=32;
    if (hc != pc) return 0;
  }
}

void divide_handler(char *from, char **heads, char **bodys)
{
  int i, j; FILE *f; char *str;
  f = divide_default;
  for (i=0; i<divide_count; i++) {
    str = divide_strings[i];
    if (divide_match(from, str)) {
      f = divide_files[i];
      goto outmsg;
    }
    for (j=0; heads[j]; j++) {
      if (divide_match(heads[j], str)) {
	f = divide_files[i];
	goto outmsg;
      }
    }
  }
outmsg:
  fprintf(f, "%s\n", from); free(from);
  for (i=0; heads[i]; i++) {
    fprintf(f, "%s\n", heads[i]); free(heads[i]);
  }
  fprintf(f,"\n");
  for (i=0; bodys[i]; i++) {
    fprintf(f, "%s\n", bodys[i]); free(bodys[i]);
  }
}

void divide(char **argv)
{
  if (argv[0]==0) usage();
  divide_default = jfopen(argv[0],"a");
  argv++;
  while (*argv) {
    divide_strings[divide_count] = *argv;
    argv++;
    if (argv[0]==0) usage();
    divide_files[divide_count] = jfopen(*argv, "a"); 
    argv++;
    divide_count++;
  }
  splitmail(stdin, divide_handler);
}

/******************************************************************************
 *
 * arc
 *
 * Open and close mail archives.
 *
 *****************************************************************************/

char  arc_gzp_name[2048];
char  arc_idx_name[2048];
char  arc_rev_name[2048];
FILE *arc_gzp;
FILE *arc_idx;
FILE *arc_rev;
int   arc_nextmsg = 0;
int   arc_lastpos = 0;

void arc_open(char *base, int update)
{
  char line[1000]; int c, max, pos; char *mode;
  struct stat arc_gzpinfo; int arc_gzpok, arc_gzperrno;
  struct stat arc_idxinfo; int arc_idxok, arc_idxerrno;
  struct stat arc_revinfo; int arc_revok, arc_reverrno;

  sprintf(arc_gzp_name, "%s.gzp", base);
  sprintf(arc_idx_name, "%s.idx", base);
  sprintf(arc_rev_name, "%s.rev", base);
  
  arc_gzpok = stat(arc_gzp_name, &arc_gzpinfo); arc_gzperrno=errno;
  arc_idxok = stat(arc_idx_name, &arc_idxinfo); arc_idxerrno=errno;
  arc_revok = stat(arc_rev_name, &arc_revinfo); arc_reverrno=errno;
  
  if ((arc_gzpok<0)&&(arc_gzperrno==ENOENT)&&
      (arc_idxok<0)&&(arc_idxerrno==ENOENT)&&
      (arc_revok<0)&&(arc_reverrno==ENOENT)&&
      update) {
    arc_idx = jfopen(arc_idx_name, "w+");
    arc_gzp = jfopen(arc_gzp_name, "w+");
    arc_rev = jfopen(arc_rev_name, "w+");
    arc_lastpos = 0;
    arc_nextmsg = 0;
    huff_file_init(arc_gzp);
    return;
  }
  
  if ((arc_gzpok>=0)&&(arc_idxok>=0)&&(arc_revok>=0)) {
    mode = update ? "r+" : "r";
    arc_idx = jfopen(arc_idx_name, mode);
    arc_gzp = jfopen(arc_gzp_name, mode);
    arc_rev = jfopen(arc_rev_name, mode);
    fseek(arc_idx, -2000, 2);
    while (c=getc(arc_idx)!='\n');
    while (1) {
      line[0]=0; pos = ftell(arc_idx);
      fgets(line, 999, arc_idx);
      if (line[0]==0) break;
      arc_lastpos = pos;
      arc_nextmsg = atoi(line)+1;
    }
    jfseek(arc_idx, 0, 2);
    fflush(arc_idx);
    return;
  }
  
  fprintf(stderr,"Cannot open archive.\n");
  exit(1);
}

void arc_close()
{
  fclose(arc_gzp);
  fclose(arc_idx);
  fclose(arc_rev);
}

void arc_error()
{
  fprintf(stderr,"Archive file corrupted.  'jmail arc-dump' will recover\n");
  fprintf(stderr,"as much information as possible.  That information can\n");
  fprintf(stderr,"then be built into a new archive.\n");
  exit(1);
}
  
/******************************************************************************
 *
 * arc_build
 *
 *****************************************************************************/

void arc_build_handler(char *from, char **heads, char **bodys)
{
  char user[100], email[100], subj[100]; int date;
  char *msg; int msglen, msgpos, msgno, i;
  
  msgno = arc_nextmsg++;
  parse_header(heads, user, email, subj, &date);
  msg = message_string(from, heads, bodys);
  msglen = strlen(msg);
  msgpos = huff_file_append(arc_gzp, msg, msglen);
  fprintf(arc_idx, "%d,%d,%d,%d,%s,%s,%s\n", 
	  msgno, msgpos, msglen, date, email, user, subj);
  if ((msgno%100)==0) printf("msg #%d\n",msgno);
  free(msg); free(from);
  for (i=0; heads[i]; i++) free(heads[i]);
  for (i=0; bodys[i]; i++) free(bodys[i]);
}

void arc_build(char **argv)
{
  if ((argv[0]==0)||(argv[1]!=0)) usage();
  arc_open(argv[0], 1);
  splitmail(stdin, arc_build_handler);
  arc_close();
}

/******************************************************************************
 *
 * arc-dump
 *
 *****************************************************************************/

void arc_dump(char **argv)
{
  int pos = 0; char buf[1024]; int nread;
  if ((argv[0]==0)||(argv[1]!=0)) usage();
  arc_open(argv[0], 0);
  while (1) {
    nread = huff_file_read(arc_gzp, pos, buf, 1024, &pos);
    if (nread > 0) fwrite(buf, 1, nread, stdout);
    if (nread != 1024) break;
  }
  fflush(stdout);
  arc_close(argv[0]);
}


/******************************************************************************
 *
 * arc-indexify
 *
 *****************************************************************************/

typedef struct word
{
  struct word *next;
  int msgcount, msgmax;
  int *msgarray;
  char text[1];
}
*word;

int   arc_indexify_minsample;
int   arc_indexify_threshold;
int   arc_indexify_minoccur;
word *arc_indexify_table;

void arc_indexify_markword(word w, int msgnum)
{
  int i;
  if ((w->msgcount>0)&&(w->msgarray[w->msgcount-1]==msgnum))
    return;
  if (w->msgcount==-1) return;
  if ((msgnum >= arc_indexify_minsample) &&
      ((w->msgcount*100)>(msgnum*arc_indexify_threshold))) {
    w->msgcount = -1;
    w->msgmax = -1;
    free(w->msgarray);
    return;
  }
  if (w->msgcount==w->msgmax) {
    int newsize = (w->msgmax << 1);
    w->msgarray = (int *)realloc(w->msgarray, newsize*sizeof(int));
    w->msgmax = newsize;
  }
  w->msgarray[w->msgcount++] = msgnum;
}

word arc_indexify_findword(char *s)
{
  word w; char *p = s;
  int hashval = strhash(s, ARC_REV_TABSIZE);
  for (w=arc_indexify_table[hashval]; w; w=w->next) {
    if (strcmp(w->text, s)==0) break;
  }
  if (w==0) {
    w = (word)malloc(sizeof(struct word)+strlen(s));
    strcpy(w->text, s);
    w->msgcount=0; w->msgmax=6;
    w->msgarray=(int *)malloc(6*sizeof(int));
    w->next = arc_indexify_table[hashval];
    arc_indexify_table[hashval] = w;
  }
  return w;
}

void arc_indexify_addwords(char *s, int msgnum)
{
  char *p, *e; char c; word w;
  while (1) {
    while (*s && !ch_alpha(*s)) s++;
    if (*s==0) break;
    e = s;
    while (*e && ch_alpha(*e)) e++;
    c = *e; *e = 0; w = arc_indexify_findword(s); *e = c;
    arc_indexify_markword(w, msgnum);
    s = e;
  }
}

void arc_indexify_dump()
{
  int prev, i, j, count, pos; word w;
  jfseek(arc_rev, 0, 0);
  ftruncate(fileno(arc_rev), 0);
  jwriteint(ARC_REV_TABSIZE, arc_rev);
  for (i=0; i<ARC_REV_TABSIZE; i++) 
    jwriteint(0, arc_rev);
  for(i=0; i<ARC_REV_TABSIZE; i++) {
    count=0; 
    for (w=arc_indexify_table[i]; w; w=w->next) {
      if ((w->msgcount>0)&&(w->msgcount<arc_indexify_minoccur)) continue;
      count++;
    }
    if (count) {
      pos = ftell(arc_rev);
      jwriteintpk(count, arc_rev);
      for (w=arc_indexify_table[i]; w; w=w->next) {
	if ((w->msgcount>0)&&(w->msgcount<arc_indexify_minoccur)) continue;
	fprintf(arc_rev, "%s ",w->text);
	if (w->msgcount<=0) jwriteintpk(0, arc_rev);
	else {
	  jwriteintpk(w->msgcount, arc_rev);
	  prev = 0;
	  for (j=0; j<w->msgcount; j++) {
	    jwriteintpk(w->msgarray[j] - prev, arc_rev);
	    prev = w->msgarray[j];
	  }
	}
      }
    } else pos=0;
    arc_indexify_table[i] = (word)pos;
  }
  jfseek(arc_rev, 4, 0);
  for (i=0; i<ARC_REV_TABSIZE; i++)
    jwriteint((int)arc_indexify_table[i], arc_rev);
}

void arc_indexify(char **argv)
{
  char line[1000]; int msgno, msgpos, msglen, nread, prevend=0;
  char *message, *firstend;
  int argc=0;
  while (argv[argc]) argc++;
  if (argc != 3) usage();
  arc_open(argv[0], 1);
  arc_indexify_threshold = atoi(argv[1]);
  arc_indexify_minoccur = atoi(argv[2]);
  arc_indexify_minsample = arc_nextmsg-1;
  if (arc_indexify_minsample > 5000) arc_indexify_minsample = 5000;
  arc_indexify_table = (word *)malloc(ARC_REV_TABSIZE * sizeof(int));
  memset(arc_indexify_table, 0, ARC_REV_TABSIZE * sizeof(int));
  jfseek(arc_idx, 0, 0);
  while (fgets(line, 999, arc_idx)) {
    msgno = atoi(line);
    msgpos = atoi(field(1, line));
    msglen = atoi(field(2, line));
    if (msgpos != prevend) arc_error();
    message = (char *)malloc(msglen+1);
    message[msglen]=0;
    nread = huff_file_read(arc_gzp, msgpos, message, msglen, &prevend);
    if (nread != msglen) arc_error();
    firstend = strchr(message, '\n');
    if (firstend==0) arc_error();
    *firstend = 0;
    if (!line_isfrom(message)) arc_error();
    *firstend = '\n';
    downcase(message);
    arc_indexify_addwords(message, msgno);
    free(message);
    if ((msgno%100)==0) printf("msg #%d\n",msgno);
  }
  arc_indexify_dump();
  arc_close();
}

/******************************************************************************
 *
 * arc-lookup
 *
 *****************************************************************************/

void display_message_text(char *line)
{
  int nread;
  int msgpos = atoi(field(1, line));
  int msglen = atoi(field(2, line));
  char *message = (char *)malloc(msglen+1);
  message[msglen] = 0;
  nread = huff_file_read(arc_gzp, msgpos, message, msglen, NULL);
  if (nread != msglen) arc_error();
  printf("%s", message);
  free(message);
}

void display_message_html(char *line)
{
  int nread; char *body;
  int msgpos = atoi(field(1, line));
  int msglen = atoi(field(2, line));
  char *message = (char *)malloc(msglen+1);
  message[msglen] = 0;
  nread = huff_file_read(arc_gzp, msgpos, message, msglen, NULL);
  if (nread != msglen) arc_error();
  body = strchr(message,'\n');
  while (body[1]!='\n') body = strchr(body+1, '\n');
  body+=2;
  
  printf("Subject: <B>");
  htmlfield(6, line);
  printf("</B><BR>\n");
  printf("Author: <B>");
  htmlfield(4, line);
  printf("</B><BR><HR>\n");
  htmlputs(body);
  printf("<HR>\n");
  printf("<I>\n");
  *body = 0;
  htmlputs(message);
  printf("</I>\n");
  free(message);
}

void display_index_text(char *line)
{
  int msgno = atoi(field(0, line));
  char *rest = field(3, line);
  printf("%d,%s",msgno,rest);
}

void display_index_html(char *line)
{
  int msgno = atoi(field(0, line));
  printf("<TR><TD><A HREF=SCRIPT?%d>%d</A></TD>",msgno,msgno);
  printf("<TD WIDTH=200>");
  htmlfield(4, line);
  printf("</TD>");
  printf("<TD WIDTH=1000>");
  htmlfield(6, line);
  printf("<BR></TD></TR>\n");
}

void display(char *line, int mode)
{
  switch(mode) {
  case 'p': display_index_text(line); break;
  case 'b': display_message_text(line); break;
  case 'P': display_index_html(line); break;
  case 'B': display_message_html(line); break;
  }
}

void display_html_table_begin()
{
  printf("<TABLE CELLSPACING=0 CELLPADDING=0>\n");
}

void display_html_table_end()
{
  printf("</TABLE>\n");
}

void display_range_msgnos(int lo, int hi, int mode)
{
  char line[1000]; double frac, fguess;
  int plo, phi, vlo, vhi, guess, c, n, pos;

  if (arc_nextmsg==0) return;
  plo = 0; phi = arc_lastpos;
  vlo = 0; vhi = arc_nextmsg - 1;
  
  while (1) {
    frac = (lo-vlo); frac/=(vhi-vlo);
    fguess = ((phi-plo)*frac) + plo;
    guess = fguess;
    if (guess > phi-1000) guess=phi-1000;
    if (guess <= plo) { guess=plo; break; }
    jfseek(arc_idx, guess, 0);
    fgets(line, 999, arc_idx);
    pos = ftell(arc_idx);
    fgets(line, 999, arc_idx);
    n = atoi(line);
    if ((n<=lo)&&(n>(lo-10))) { guess=pos; break; }
    if (n>lo) { phi = pos; vhi = n; }
    else { plo = pos; vlo = n; }
  }
  jfseek(arc_idx, guess, 0);
  while(1) {
    line[0]=0;
    fgets(line, 999, arc_idx);
    if (line[0]==0) break;
    n = atoi(line);
    if (n>hi) break;
    if (n>=lo) display(line, mode);
  }
}

void display_range_dates(int lo, int hi, int mode)
{
  char line[1000];
  int plo, phi, mid, n, pos;
  plo = 0; phi = arc_lastpos;
  while ((phi-2000) > plo) {
    mid = (phi-1000+plo)>>1;
    jfseek(arc_idx, mid, 0);
    fgets(line, 999, arc_idx);
    pos = ftell(arc_idx);
    while (1) {
      fgets(line, 999, arc_idx);
      n = atoi(field(3, line));
      if (n) break;
    }
    if (pos==phi) phi=pos;
    else if (n>(lo-260000)) phi=pos;
    else plo=pos;
  }
  jfseek(arc_idx, plo, 0);
  while(1) {
    line[0]=0;
    fgets(line, 999, arc_idx);
    if (line[0]==0) break;
    n = atoi(field(3, line));
    if (n>hi+260000) break;
    if ((n>=lo)&&(n<=hi)) display(line, mode);
  }
}

intset searchword(char *s)
{
  char buf[1000]; char *p = s; intset res;
  int nbuckets, hashval, prev, i, j, ok, bucketpos, bucketwords, wordrefs;

  jfseek(arc_rev, 0, 0);
  nbuckets = jreadint(arc_rev);
  hashval = strhash(s, nbuckets);
  jfseek(arc_rev, (hashval*4)+4, 0);
  bucketpos = jreadint(arc_rev);
  if (bucketpos == 0) return INTSET_NONE;
  jfseek(arc_rev, bucketpos, 0);
  bucketwords = jreadintpk(arc_rev);
  for (i=0; i<bucketwords; i++) {
    jreadword(arc_rev, buf, 999);
    wordrefs = jreadintpk(arc_rev);
    if (strcmp(buf, s)==0) {
      if (wordrefs==0) return INTSET_ALL;
      res = intset_make(wordrefs);
      prev=0;
      for (j=0; j<wordrefs; j++) {
	prev += jreadintpk(arc_rev);
	res->data[j] = prev;
      }
      return res;
    } else {
      for (j=0; j<wordrefs; j++)
	jreadintpk(arc_rev);
    }
  }
  return INTSET_NONE;
}

void display_word_count(char *w, intset v, int mode)
{
  switch(mode) {
  case 'p':
    if (v==INTSET_ALL) 
      printf("# The word %s occurs in many messages.\n", w);
    else
      printf("# The word %s occurs in %d messages.\n", w, v->size);
    break;
  case 'P':
    if (v==INTSET_ALL) 
      printf("The word <b>%s</b> occurs in <b>many</b> messages.<br>\n", w);
    else
      printf("The word <b>%s</b> occurs in <b>%d</b> messages.<br>\n",
	     w, v->size);
    break;
  }
}

void display_find_words(char *wordstr, int mode)
{
  int nwords, i;
  char *words[1000];
  intset vectors[1000];
  char *p, *e; intset finalunion;

  /* extract the words */
  nwords = 0;
  p = wordstr;
  while (1) {
    while (*p && (!ch_alpha(*p))) p++;
    if (*p == 0) break; e = p;
    while (ch_alpha(*e)) e++;
    words[nwords] = strclip(p, e);
    downcase(words[nwords]);
    nwords++; p = e;
  }
  /* obtain the word-bitvectors */
  if (mode=='P') printf("<HR>\n");
  for (i=0; i<nwords; i++) {
    vectors[i] = searchword(words[i]);
    display_word_count(words[i], vectors[i], mode);
  }
  if (mode=='P') printf("<HR>\n");
  finalunion = intset_union(vectors, nwords);
  if (finalunion == INTSET_ALL) return;
  if (mode=='P') display_html_table_begin();
  for (i=0; i<finalunion->size; i++) {
    display_range_msgnos(finalunion->data[i], finalunion->data[i], mode);
  }
  if (mode=='P') display_html_table_end();
  intset_free(finalunion);
  for (i=0; i<nwords; i++) {
    intset_free(vectors[i]);
    free(words[i]);
  }
}

void arc_lookup(char **argv, int dmode)
{
  char line[1000];
  int lo, hi;
  if ((argv[0]==0)||(argv[1]!=0)) usage();
  arc_open(argv[0], 0);
  while (fgets(line, 999, stdin)) {
    switch (line[0]) {
    case 'n':
      sscanf(line+1,"%d%d",&lo,&hi);      
      if (dmode=='P') display_html_table_begin();
      if ((lo<=hi)&&(lo>=0)||(hi>=0))
	display_range_msgnos(lo,hi,dmode);
      if (dmode=='P') display_html_table_end();
      break;
    case 'd':
      sscanf(line+1,"%d%d",&lo,&hi);
      if (dmode=='P') display_html_table_begin();
      if ((lo<=hi)&&(lo>=0)||(hi>=0))
	display_range_dates(lo,hi,dmode);
      if (dmode=='P') display_html_table_end();
      break;
    case 'l':
      sscanf(line+1,"%d",&lo); 
      hi=time(0); lo=hi-lo; hi+=172800;
      if (lo<0) lo=0;
      if (dmode=='P') display_html_table_begin();
      display_range_dates(lo, hi, dmode);
      if (dmode=='P') display_html_table_end();
      break;
    case 'w':
      display_find_words(line+1, dmode);
      break;
    default:
      sscanf(line,"%d",&lo);
      if (lo>=0) display_range_msgnos(lo, lo, dmode);
      break;
    }
  }
}

/******************************************************************************
 *
 * unmangle
 *
 *****************************************************************************/

char unmangle_fromline[10000];

void unmangle_printmsg(char **lines, int nlines)
{
  int i = 0, j = 0; int havefrom=0, havesubj=0; char *line;
  if (nlines==0) return;
  if (line_isfrom(lines[i])) {
    strcpy(unmangle_fromline, lines[i]);
    i++;
  }
  printf("%s\n", unmangle_fromline);
  while ((i<nlines) && (!line_isblank(lines[i]))) {
    line = lines[i];
    if (!line_isrfc(line)) {
      while ((*line)&&(*line<=32)) line++;
      printf("Whatever: %s\n", lines[i]);
    } else printf("%s\n", lines[i]);
    if (strncasecmp(lines[i], "From:", 5)==0) havefrom=1;
    if (strncasecmp(lines[i], "Subject:", 8)==0) havesubj=1;
    i++;
  }
  if (!havefrom) printf("From: anonymous\n");
  if (!havesubj) printf("Subject: none\n");
  printf("\n");
  j = nlines;
  while ((i<j) && (line_isnull(lines[i]))) i++;
  while ((j>i) && (line_isnull(lines[j-1]))) j--;
  while (i<j) { printf("%s\n", lines[i]); i++; }
  printf("\n");
  for (i=0; i<nlines; i++) free(lines[i]);
}

int unmangle_have_head(char **lines, int nlines)
{
  int i, front = nlines; int havefrom=0, havesubj=0;
  front--; if (front < 0) return 0;
  if (!line_isblank(lines[front])) return 0;
  front--; if (front < 0) return 0;
  if (!line_isrfc(lines[front])) return 0;
  front--; if (front < 0) return 0;
  if (!line_isrfc(lines[front])) return 0;
  front--;
  while ((front>=0)&&(line_isrfc(lines[front]))) front--;
  if ((front>=0)&&(line_isfrom(lines[front]))) front--;
  front++;
  for (i=front; i<nlines; i++) {
    if (strncasecmp(lines[i], "From:", 5)==0) havefrom=1;
    if (strncasecmp(lines[i], "Subject:", 8)==0) havesubj=1;
  }
  if ((havefrom==0)&&(havesubj==0)) return 0;
  return front;
}

void unmangle(char **argv)
{
  char line[10000];
  char **lines, *p; int nlines, i, headp;
  if (argv[0]) usage();
  lines = (char **)malloc(1000000*sizeof(char *));
  nlines = 0;
  strcpy(unmangle_fromline, "From nobody@nowhere  Mon Jan  1 00:00:00 1900");
  while (fgets(line, 9999, stdin)) {
    p=line+strlen(line);
    while ((p>line)&&(p[-1]<=32)) p--;
    p[0]=0;
    if (line_isfrom(line)) {
      unmangle_printmsg(lines, nlines);
      nlines = 0;
    }
    lines[nlines++] = strdup(line);
    headp = unmangle_have_head(lines, nlines);
    if (headp) {
      unmangle_printmsg(lines, headp);
      nlines -= headp;
      for (i=0; i<nlines; i++) lines[i]=lines[i+headp];
    }
  }
  if (nlines) unmangle_printmsg(lines, nlines);
}

/******************************************************************************
 *
 * claimlock
 *
 *****************************************************************************/

void claimlock(char **argv)
{
  int fd;
  if ((argv[0]==0)||(argv[1]!=0)) usage();
  while (1) {
    fd = open(argv[0], O_RDWR | O_CREAT | O_EXCL, 0777);
    if (fd >= 0) break;
    if (errno != EEXIST) {
      fprintf(stderr,"claim-lockfile: %s\n", strerror(errno));
      exit(1);
    }
    sleep(1);
  }
  exit(0);
}

/******************************************************************************
 *
 * main
 *
 *****************************************************************************/

usage()
{
  fprintf(stderr,"usage: jmail unmangle < in > out\n");
  fprintf(stderr,"usage: jmail keep-headers head1 head2  ... < in > out\n");
  fprintf(stderr,"usage: jmail divide file0 string1 file1 string2 file2... < in\n");
  fprintf(stderr,"usage: jmail claim-lockfile filename\n");
  fprintf(stderr,"usage: jmail arc-build arcname < in\n");
  fprintf(stderr,"usage: jmail arc-dump arcname > out\n");
  fprintf(stderr,"usage: jmail arc-indexify arcname prefix threshold minoccur < in\n");
  fprintf(stderr,"usage: jmail arc-lookup-text arcname < cmds > out\n");
  fprintf(stderr,"usage: jmail arc-lookup-html arcname < cmds > out\n");
  fprintf(stderr,"usage: jmail arc-lookup-text-index arcname < cmds > out\n");
  fprintf(stderr,"usage: jmail arc-lookup-html-index arcname < cmds > out\n");
  exit(1);
}

main(int argc, char **argv)
{
  if (argc<2) usage();
  if (strcmp(argv[1],"keep-headers")==0) { keep_headers(argv+2); exit(0); }
  if (strcmp(argv[1],"divide")==0) { divide(argv+2); exit(0); }
  if (strcmp(argv[1],"arc-build")==0) { arc_build(argv+2); exit(0); }
  if (strcmp(argv[1],"arc-dump")==0) { arc_dump(argv+2); exit(0); }
  if (strcmp(argv[1],"arc-indexify")==0) { arc_indexify(argv+2); exit(0); }
  if (strcmp(argv[1],"arc-lookup-text-index")==0) {arc_lookup(argv+2,'p');exit(0);}
  if (strcmp(argv[1],"arc-lookup-text")==0) {arc_lookup(argv+2,'b');exit(0);}
  if (strcmp(argv[1],"arc-lookup-html-index")==0) {arc_lookup(argv+2,'P');exit(0);}
  if (strcmp(argv[1],"arc-lookup-html")==0) {arc_lookup(argv+2,'B');exit(0);}
  if (strcmp(argv[1],"unmangle")==0) { unmangle(argv+2); exit(0); }
  if (strcmp(argv[1],"claim-lockfile")==0) { claimlock(argv+2); exit(0); }
  usage();
}
